﻿//--------------------------------------------------------------
//
// Eamon Revolutions Role Playing Game
// Derek C. Jeter
//
//--------------------------------------------------------------

function XGameEngine()
{
    var self = this;

    self.GAME_ENGINE_MAJOR_VERSION = "1";
    self.GAME_ENGINE_MINOR_VERSION = "0";

    self.context = {};
    self.commandHandlers = {};

    self.initialize = function(consoleImpl, pmImpl, state, rndNum, cbtEngine)
    {
        self.context.console = consoleImpl;
        self.context.persistenceManager = pmImpl;
        self.context.gameStateData = state;
        self.context.gameEngine = self;
        self.context.rndNum = rndNum;
        self.context.combatEngine = cbtEngine;

        self.registerCommandHandlers();
        self.initGameModule();
        self.context.combatEngine.initialize(self.context);

        consoleImpl.log("GAME ENGINE INITIALIZED");
    };

    self.startup = function()
    {
        var cnsl = self.context.console;

        cnsl.writeLine("Welcome to Eamon on the Web!", "yellow");
        cnsl.writeLine("Programmed by Derek C. Jeter", "yellow");
        cnsl.writeLine("Based on Eamon Deluxe by Frank Black", "yellow");
        cnsl.writeLine("Game Engine Version " + self.GAME_ENGINE_MAJOR_VERSION + "." + self.GAME_ENGINE_MINOR_VERSION);
        cnsl.writeBlankLine();
        cnsl.writeLine("You are playing '" + self.context.gameStateData.gameModule.name + "'.", "lightgreen");
        if (!_.isUndefined(self.context.gameStateData.gameModule.info && !_.isUndefined(self.context.gameStateData.gameModule.info.author)))
        {
            cnsl.writeLine("By " + self.context.gameStateData.gameModule.info.author, "lightgreen");
        }
        cnsl.writeBlankLine();
        cnsl.writeBlankLine();

        if (self.context.gameStateData.round == 1)
        {
            self.printIntro();
        }

        cnsl.flush();

        this.beginTurn(CmdResultFactory.okVerbose());
    };

    self.processCommand = function(commandText)
    {
        var cnsl = self.context.console;

        cnsl.writeLine(commandText); // Echo the command
        cnsl.writeBlankLine();

        var pc = this.parseCommand(commandText.trim());

        var cmdHandler = self.commandHandlers[pc.verb];

        if (_.isUndefined(cmdHandler))
        {
            cnsl.writeLine("Huh? I don't understand that command!", "red");
            //TODO: Dump out a list of available commands or implement a help command?

            var r = CmdResultFactory.error();

            self.beginTurn(r);

            return r;
        }

        if (cmdHandler.requiresDebugMode && !this.context.gameStateData.debugMode)
        {
            cnsl.writeLine("ILLEGAL COMMAND - Nice try, though!", "red");

            var x = CmdResultFactory.error();

            this.beginTurn(x);

            return x;
        }

        if (!cmdHandler.canUseWhenDead && this.context.gameStateData.charIsDead)
        {
            cnsl.writeLine("You are dead! Try using QUIT or RESTORE.", "red");

            var y = CmdResultFactory.error();

            this.beginTurn(y);

            return y;
        }

        var cmdResults;
        if (self.context.gameStateData.debugMode)
        {
            // Let errors bubble up in debug mode so they are logged by browser console
            cmdResults = cmdHandler.execute(self.context, pc);
        }
        else
        {
            try
            {
                cmdResults = cmdHandler.execute(self.context, pc);
            }
            catch (err)
            {
                cnsl.writeLine("ERROR: " + err.message + " (Line: " + err.lineNumber + ")", "red");

                var e = CmdResultFactory.error();

                this.beginTurn(e);

                return e;
            }
        }

        this.beginTurn(cmdResults);

        return cmdResults;
    };

    self.parseCommand = function(commandText)
    {
        var parsedCommand = new Object();

        var index = commandText.indexOf(' ');
        var params = '';

        if (index == -1)
        {
            parsedCommand.verb = commandText.trim().toLowerCase();
        }
        else
        {
            parsedCommand.verb = commandText.substring(0, index).trim().toLowerCase();
            params = commandText.substring(index + 1).trim().toLowerCase();
        }

        parsedCommand.directObject = params;
        parsedCommand.indirectObject = '';
        parsedCommand.preposition = '';
        
        var preps = ['in', 'from', 'at', 'to', 'into', 'with', 'on'];

        for (var i = 0; i < preps.length; i++) {
            index = params.indexOf(' ' + preps[i] + ' ');

            if (index > -1) {
                parsedCommand.directObject = params.substring(0, index).trim();
                parsedCommand.preposition = params.substring(index + 1, index + preps[i].length + 1).trim();
                parsedCommand.indirectObject = params.substring(index + preps[i].length + 1).trim();

                break;
            }
        }

        return parsedCommand;
    };

    self.beginTurn = function(cmdResults)
    {
        var cnsl = self.context.console;
        var state = self.context.gameStateData;

        cnsl.log("*** Room # = " + state.room + " [Round #" + state.round + "]");

        if (!cmdResults.inhibitDescription)
        {
            self.describeRoom();
        }
        else
        {
            cnsl.writeBlankLine();
        }

        // Monsters do their thang
        if (!cmdResults.inhibitEncounter)
        {
            self.monstersTakeAction();
        }

        if (!cmdResults.inhibitRoundAdvance)
        {
            state.round++;

            self.updateCounters();
        }

        cnsl.write("Your Command?&nbsp;&nbsp;", "lightBlue");

        cnsl.flush();
    };

    self.describeRoom = function()
    {
        var cnsl = self.context.console;
        var state = self.context.gameStateData;

        cnsl.writeRuleLine();

        var room = self.getRoom(state.room);

        // Check for dark room/no light source
        if (room.isDark && state.lightSource == 0)
        {
            cnsl.writeLine("It's too dark to see.");

            return;
        }

        // Warn of wasting light
        if (!room.isDark && state.lightSource > 0)
        {
            var artifact = self.getArtifact(state.lightSource);

            if (artifact.lightCounter != -1) // Not an infinite light source
            {
                cnsl.writeLine("It's not dark here. You might want to EXTINGUISH your " + artifact.name + ".");
            }
        }

        // Room Name
        cnsl.writeLine(room.name, "yellow");
        cnsl.writeBlankLine();

        // Room Description
        if (!room.hasBeenSeen)
        {
            self.printDescription(room.description, "white");
            cnsl.writeBlankLine();

            room.hasBeenSeen = true;
        }

        // Monsters in room
        _.each(self.context.gameStateData.gameModule.monsters, function(monster)
        {
            if (monster.locationDetail.type == CLocationType.room && monster.locationDetail.number == room.number)
            {
                if (!monster.hasBeenSeen)
                {
                    self.printDescription(monster.description, "white");
                    cnsl.writeBlankLine();

                    if (monster.personality > CPersonalityType.max)
                    {
                        // Determine personality based on charisma-adjusted roll
                        var basePct = monster.personality - 100;
                        var charismaAdjustPct = (self.context.gameStateData.runChar.charisma - 10) * 2;

                        if (self.context.rndNum.checkPct(basePct + charismaAdjustPct))
                        {
                            monster.personality = CPersonalityType.friend;
                        }
                        else
                        {
                            monster.personality = CPersonalityType.enemy;
                        }

                        cnsl.log("Monster '" + monster.name + "' is personality = " + monster.personality + "; Base Pct = " + basePct + ";  Chr adj = " + charismaAdjustPct);
                    }

                    monster.hasBeenSeen = true;
                }

                var x = (monster.currentGroupSize > 1) ? "are" : "is";
                cnsl.writeLine(self.getMonsterName(monster) + " " + x + " here.");
                cnsl.writeBlankLine();
            }
        });

        // Artifacts in room
        _.each(self.context.gameStateData.gameModule.artifacts, function(a)
        {
            if (a.locationDetail.type == CLocationType.room && a.locationDetail.number == room.number)
            {
                //TODO: Check if the artifact came from the character's weapon list, if so, print "Your <item> is here."

                if (!a.hasBeenSeen)
                {
                    self.printDescription(a.description, "white");
                    cnsl.writeBlankLine();

                    a.hasBeenSeen = true;
                }

                cnsl.writeLine("You see " + a.name + ".");
                cnsl.writeBlankLine();
            }
        });
    };

    self.updateCounters = function()
    {
        var cnsl = self.context.console;
        var state = self.context.gameStateData;

        // Check light counter
        if (state.lightSource > 0)
        {
            var artifact = self.getArtifact(state.lightSource);

            if (artifact.lightCounter != -1)
            {
                // Not an infinite light source, decrement counter
                artifact.lightCounter--;

                if (artifact.lightCounter < 1)
                {
                    cnsl.writeLine("The " + artifact.name + " goes out.");
                    cnsl.writeBlankLine();

                    state.lightSource = 0;
                }
                else if (artifact.lightCounter < 10)
                {
                    cnsl.writeLine("The " + artifact.name + " grows dim!");
                    cnsl.writeBlankLine();
                }
                else if (artifact.lightCounter < 20)
                {
                    cnsl.writeLine("The " + artifact.name + " is almost out!");
                    cnsl.writeBlankLine();
                }

                cnsl.log("Remaining light: " + artifact.lightCounter);
            }
        }

        // Update speed counter
        if (self.context.gameStateData.speedCounter > 0)
        {
            self.context.gameStateData.speedCounter--;

            cnsl.log("Remaining speed: " + self.context.gameStateData.speedCounter);

            if (self.context.gameStateData.speedCounter == 0)
            {
                self.context.gameStateData.runChar.agility = self.context.gameStateData.initChar.agility;

                cnsl.writeLine("Your speed spell has expired!");
            }
        }

        // Recharge spells
        _.each(_.keys(state.initChar.spellAbilities), function(spell)
        {
            var originalAbility = state.initChar.spellAbilities[spell];
            var currentAbility = state.runChar.spellAbilities[spell];

            if (currentAbility < originalAbility)
            {
                state.runChar.spellAbilities[spell] = currentAbility + 1;

                cnsl.log(spell.toUpperCase() + " recharge: " + state.runChar.spellAbilities[spell] + " of " + originalAbility);
            }
        });
    };

    self.monstersTakeAction = function()
    {
        var monsters = _.filter(self.context.gameStateData.gameModule.monsters, function(x)
        {
            // Get all monsters in room
            return (x.locationDetail.type == CLocationType.room && x.locationDetail.number == self.context.gameStateData.room);
        });

        _.each(monsters, function(m)
        {
            //TODO: monster attacks or flees
        });
        
    };

    self.getRoom = function(n)
    {
        var x = _.findWhere(self.context.gameStateData.gameModule.rooms, { number: n });

        if (_.isUndefined(x))
        {
            self.context.console.writeLine("[ERROR] Invalid Room #" + n, "red");
            return null;
        }

        return x;
    };

    self.getArtifact = function(n)
    {
        var x = _.findWhere(self.context.gameStateData.gameModule.artifacts, { number: n });

        if (_.isUndefined(x)) {
            self.context.console.writeLine("[ERROR] Invalid Artifact #" + n, "red");
            return null;
        }

        return x;
    };

    self.getMonster = function(n)
    {
        var x = _.findWhere(self.context.gameStateData.gameModule.monsters, { number: n });

        if (_.isUndefined(x)) {
            self.context.console.writeLine("[ERROR] Invalid Monster #" + n, "red");
            return null;
        }

        return x;
    };

    self.getMonsterName = function(m)
    {
        var name = m.name;

        if (m.number == 0) // Player
        {
            return name;
        }

        if (m.currentGroupSize > 1)
        {
            name += "s";
        }

        return name;
    };

    self.getTargetNamePossesive = function(target)
    {
        var n = target.name;

        if (n.toLowerCase()[n.length - 1] == 's')
        {
            n += "'";
        }
        else
        {
            n += "'s";
        }

        return n;
    }

    // searchString - what to look for, usually the subject of a command
    // whereToLook - array of CLocationType to narrow search
    self.artifactSearch = function(searchString, whereToLook, number)
    {
        var sl = searchString.length;
        var results = [];

        _.each(self.context.gameStateData.gameModule.artifacts, function(x)
        {
            // Narrow search by location
            if (_.indexOf(whereToLook, x.locationDetail.type) < 0)
            {
                return;
            }

            // Verify room number
            if ((x.locationDetail.type == CLocationType.room || x.locationDetail.type == CLocationType.embedded) && x.locationDetail.number != self.context.gameStateData.room)
            {
                return;
            }

            // Verify monster number
            if ((x.locationDetail.type == CLocationType.carriedByMonster || x.locationDetail.type == CLocationType.wornByMonster) && x.locationDetail.number != number)
            {
                return;
            }

            // Verify container number
            if (x.locationDetail.type == CLocationType.containedInArtifact && x.locationDetail.number != number)
            {
                return;
            }

            // Check for name match
            if (x.name.toLowerCase().indexOf(searchString) > -1)
            {
                results.push(x);
            }
            else if (_.isArray(x.synonyms))
            {
                // Check for synonym match
                for (var i = 0; i < x.synonyms.length; i++)
                {
                    var s = x.synonyms[i].toLowerCase();

                    if (s.substr(0, sl) == searchString || s.substr(s.length - sl, sl) == searchString)
                    {
                        results.push(x);
                    }
                }
            }
        });

        return results;
    };

    self.typedArtifactSearch = function(cmd, artifactTypes, whereToLook)
    {
        var matches = self.context.gameEngine.artifactSearch(cmd.directObject, whereToLook);

        if (matches.length == 0)
        {
            self.context.console.writeLine("You can't " + cmd.verb + " that.");
            return null;
        }

        if (matches.length > 1)
        {
            self.context.console.writeLine("Please be more specific!");
            return null;
        }

        if (_.indexOf(artifactTypes, matches[0].type) < 0)
        {
            self.context.console.writeLine("You can't " + cmd.verb + " that.");
            return null;
        }

        return matches[0];
    }

    self.inventoryArtifactSearch = function(cmd, artifactTypes)
    {
        var matches = self.context.gameEngine.artifactSearch(cmd.directObject, [CLocationType.carriedByPlayer, CLocationType.wornByPlayer]);

        if (matches.length == 0)
        {
            self.context.console.writeLine("You can't " + cmd.verb + " that because you don't have it.");
            return null;
        }

        if (matches.length > 1)
        {
            self.context.console.writeLine("Please be more specific!");
            return null;
        }

        if (!_.isUndefined(artifactTypes) && artifactTypes != null)
        {
            if (_.indexOf(artifactTypes, matches[0].type) < 0)
            {
                self.context.console.writeLine("You can't " + cmd.verb + " that.");
                return null;
            }
        }

        return matches[0];
    }

    self.monsterSearch = function(searchString)
    {
        var results = _.filter(self.context.gameStateData.gameModule.monsters, function(x)
        {
            return (x.locationDetail.type == CLocationType.room && x.locationDetail.number == self.context.gameStateData.room &&
                x.name.toLowerCase().indexOf(searchString) > -1);
        });

        if (results.length == 0)
        {
            self.context.console.writeLine("Nobody here by that name!");
            return null;
        }

        if (results.length > 1)
        {
            self.context.console.writeLine("Please be more specific!");
            return null;
        }

        return results[0];
    };

    self.targetSearch = function(cmd)
    {
        // Search monsters
        var results = _.filter(self.context.gameStateData.gameModule.monsters, function(x)
        {
            return (x.locationDetail.type == CLocationType.room && x.locationDetail.number == self.context.gameStateData.room &&
                x.name.toLowerCase().indexOf(cmd.directObject) > -1);
        });

        if (results.length == 1)
        {
            return {
                type: CTargetType.monster,
                target: results[0]
            };
        }

        if (results.length > 1)
        {
            self.context.console.writeLine("Please be more specific!");
            return null;
        }

        // Search artifacts
        var whereToLook = [CLocationType.carriedByPlayer, CLocationType.wornByPlayer, CLocationType.room, CLocationType.embedded];
        results = self.artifactSearch(cmd.directObject, whereToLook);

        if (results.length == 0)
        {
            self.context.console.writeLine("I don't understand who or what you are trying to " + cmd.verb + ".");
            return null;
        }

        if (results.length > 1)
        {
            self.context.console.writeLine("Please be more specific!");
            return null;
        }

        return {
            type: CTargetType.artifact,
            target: results[0]
        };
    }

    self.calculateEncumberance = function()
    {
        var wt = 0;

        _.each(self.context.gameStateData.gameModule.artifacts, function(a)
        {
            if (a.weight == -999)
            {
                return;
            }

            if (a.locationDetail.type == CLocationType.carriedByPlayer || a.locationDetail.type == CLocationType.wornByPlayer)
            {
                wt += a.weight;
            }
            else if (a.locationDetail.type == CLocationType.containedInArtifact)
            {
                // Also need to count weight of items in containers carried by player    
                var container = self.getArtifact(a.locationDetail.number);

                if (container.locationDetail.type == CLocationType.carriedByPlayer || container.locationDetail.type == CLocationType.wornByPlayer)
                {
                    wt += a.weight;
                }
            }
        });

        return wt;
    }

	self.printDescription = function(desc, color, includeLineBreak)
	{
	    if (_.isUndefined(includeLineBreak))
	    {
	        includeLineBreak = true;
	    }

		var l = desc.length;
		
		if (desc.charAt(l - 4) == '{' && desc.charAt(l - 1) == '}')
		{
			var colorCode = desc.substr(l - 3, 2);
		    
			color = CColorMapper.map(colorCode);

		    desc = desc.substr(0, l - 4);
		}
		
		if (includeLineBreak)
		{
		    self.context.console.writeLine(desc, color);
		}
		else
		{
		    self.context.console.write(desc, color);
		}
	};
	
	self.printEffect = function(n)
	{
		var x = _.findWhere(self.context.gameStateData.gameModule.effects, { number: n });

        if (_.isUndefined(x)) {
            self.context.console.writeLine("[ERROR] Invalid Effect #" + n, "red");
            return;
        }
		
	    // Check for continuation markers (*001 and **001)
	    var l = x.description.length;
        if (l > 5 && x.description.charAt(l - 4) == '*')
        {
            var offset = 4;
            var includeLineBreak = true;

            if (x.description.charAt(l - 5) == '*')
            {
                offset = 5;
                includeLineBreak = false;
            }

            var desc = x.description.substr(0, l - offset);

            self.printDescription(desc, "green", includeLineBreak);

            // Print linked desc
            var linkedEffect = x.description.substr(l - 3, 3);
            var linkedNumber = parseInt(linkedEffect);

            if (_.isNaN(linkedNumber))
            {
                self.context.console.writeLine("[ERROR] Invalid linked Effect #" + linkedEffect + " in effect #" + n, "red");
                return;
            }

            self.printEffect(linkedNumber);

            return;
        }
		
		self.printDescription(x.description, "green");
	};
	
	self.printEffectSequence = function(n, numEffects)
	{
		for (var i = 0; i < numEffects; i++)
		{
			self.printEffect(n + i);
		}
	};

    self.checkExit = function(direction)
    {
        var room = self.context.gameEngine.getRoom(self.context.gameStateData.room);
        var exit = _.findWhere(room.exits, { direction: direction });
        var cnsl = self.context.console;

        if (_.isUndefined(exit))
        {
            cnsl.writeLine("You can't go that way!");

            return null;
        }

        if (exit.isDoor)
        {
            var door = self.context.gameEngine.getArtifact(exit.doorArtifactNumber);

            if (door.isHidden)
            {
                cnsl.writeLine("You can't go that way!");

                return null;
            }

            if (door.openStatus == COpenStatus.closed)
            {
                cnsl.writeLine("The " + door.name + " blocks the way!");

                return null;
            }

            return door.roomBeyond;
        }

        return exit.connection;
    };

    self.checkKey = function(artifact, cmd)
    {
        if (artifact.keyNumber <= 0)
        {
            return true; // No key required
        }

        var results = self.artifactSearch(cmd.indirectObject, [CLocationType.carriedByPlayer, CLocationType.wornByPlayer, CLocationType.room]);

        if (results.length == 0)
        {
            self.context.console.writeLine("You can't " + cmd.verb + " the " + artifact.name + " with that because you don't have it.");

            return false;
        }

        if (results.length > 1)
        {
            self.context.console.writeLine("Please be more specific!");

            return false;
        }

        var key = results[0];

        if (key.number != artifact.keyNumber)
        {
            self.context.console.writeLine("You cannot " + cmd.verb + " " + artifact.name + " with that.");

            return false;
        }

        return true;
    };

    self.moveMonsters = function(connection)
    {
        var monsters = _.filter(self.context.gameStateData.gameModule.monsters, function(x)
        {
            // Get all monsters in room
            return (x.locationDetail.type == CLocationType.room && x.locationDetail.number == self.context.gameStateData.room);
        });

        _.each(monsters, function(m)
        {
            if (m.personality == CPersonalityType.friend)
            {
                // Move tag-alongs
                m.locationDetail.number = connection;
            }
            else if (m.personality == CPersonalityType.enemy)
            {
                // Check enemy pursuit
                if (self.context.rndNum.checkPct(m.courage))
                {
                    m.locationDetail.number = connection;

                    self.context.console.writeLine(self.getMonsterName(m) + " pursues!"); //TODO: Check grammar on this, esp group monsters
                }
            }
        });
    };

    self.getArmorDetails = function() //TODO: Overhaul the whole armor system
    {
        var details = {
            shieldNumber: 0,
            armorNumber: 0,
            armorType: 0,
            armorClass: 0,
            description: ""
        };

        var list = [];

        _.each(self.context.gameStateData.gameModule.artifacts, function(a)
        {
            if (a.type == CArtifactType.wearable && a.locationDetail.type == CLocationType.wornByPlayer && a.armorClass > 0)
            {
                if (a.armorType == CArmorType.shield) // Shield
                {
                    details.shieldNumber = a.number;
                    details.armorClass += a.armorClass;
                    list.push(a.name);
                }
                else // Armor
                {
                    details.armorNumber = a.number;
                    details.armorType = a.armorType;
                    details.armorClass += a.armorClass;
                    list.push(a.name);
                }
            }
        });

        details.description = list.join(", ");

        return details;
    };

    self.revealDisguisedMonster = function(artifact)
    {
        var x = (artifact.name.endsWithS()) ? "are" : "is";
        self.context.console.writeLine(artifact.name + " " + x + " alive!", "white");
        self.context.console.writeBlankLine();

        self.printEffectSequence(artifact.effectNumber, artifact.numEffects);

        artifact.locationDetail.type = CLocationType.hidden;
        artifact.locationDetail.number = 0;

        var monster = self.getMonster(artifact.disguisedMonsterNumber);

        monster.locationDetail.type = CLocationType.room;
        monster.locationDetail.number = self.context.gameStateData.room;
    };

    self.printHealthStatus = function(target, isCombatMode)
    {
        var prefix;

        if (target.number == 0)
        {
            prefix = target.name +  " is ";
        }
        else
        {
            if (target.currentGroupSize > 1)
            {
                if (isCombatMode)
                {
                    prefix = "One " + target.name + " is ";
                }
                else
                {
                    prefix = "They are ";
                }
            }
            else
            {
                prefix = target.name.capitalize() + " is ";
            }
        }

        if (target.damageTaken >= target.hardiness)
        {
            // Char/monster is dead
            self.context.console.writeLine(prefix + "dead!");

            return true;
        }

        var x = (Math.round(target.damageTaken * 5 / target.hardiness) + 1) * (target.damageTaken > 0);

        if (x < 1)
        {
            self.context.console.writeLine(prefix + "in perfect health.");

            return false;
        }

        switch(x)
        {
            case 4:
                self.context.console.writeLine(prefix + "badly injured.");
                break;

            case 3:
                self.context.console.writeLine(prefix + "in pain.");
                break;

            case 2:
                self.context.console.writeLine(prefix + "hurting.");
                break;

            case 1:
                self.context.console.writeLine(prefix + "still in good shape.");
                break;

            default:
                self.context.console.writeLine(prefix + "at death's door, knocking loudly.");
                break;
        }

        return false;
    };

    self.adjustDamage = function(target, amount)
    {
        target.damageTaken += amount;
        target.damageTaken = target.damageTaken.clamp(0, target.hardiness);

        if (!self.printHealthStatus(target, true))
        {
            return false;
        }

        // Char is dead
        if (target.number == 0)
        {
            self.context.console.writeBlankLine();
            self.context.console.writeLine("You may now only use the QUIT and RESTORE commands.", "red");
        }
        else // Monster is dead
        {
            // Remove from room
            target.locationDetail.number = 0;
            target.locationDetail.type = CLocationType.hidden;

            // If there is a dead body for this monster, place in room
            if (target.deadBodyNumber > 0)
            {
                var a = self.getArtifact(target.deadBodyNumber);

                a.locationDetail.number = self.context.gameStateData.room;
                a.locationDetail.type = CLocationType.room;
            }
        }

        return true;
    };

    self.checkSpellCast = function(spell)
    {
        var cnsl = self.context.console; 
        var ability = self.gameStateData.runChar.spellAbilities[spell];

        if (_.isUndefined(ability) || ability == 0)
        {
            cnsl.writeLine("You do not know how to cast '" + spell.toUpperCase() + "'.");
            return false;
        }

        var rl = self.context.rndNum.roll(1, 100);

        if (rl == 100)
        {
            cnsl.writeLine("The strain of attempting to cast " + spell.toUpperCase() + " overloads your brain and you forget it completely for the rest of this adventure.");

            self.gameStateData.runChar.spellAbilities[spell] = 0;

            return false;
        }

        // Ability to cast this spell is halved
        self.gameStateData.runChar.spellAbilities[spell] = Math.round(self.gameStateData.runChar.spellAbilities[spell] / 2);

        if (rl < ability || rl < 5)
        {
            return true;
        }

        return false;
    };

    self.checkSpellAbilityIncrease = function(spell)
    {
        var ability = self.gameStateData.runChar.spellAbilities[spell];

        if (!self.context.rndNum.checkPct(ability))
        {
            self.gameStateData.runChar.spellAbilities[spell] = ability + 2;

            self.context.console.writeLine("Your ability to cast " + spell.toUpperCase() + " just increased!");
        }
    };

    self.checkItemGet = function(artifact)
    {
        var cnsl = self.context.console;

        // Check for disguised monster
        if (artifact.type == CArtifactType.disguisedMonster)
        {
            self.revealDisguisedMonster(artifact);

            return false;
        }

        // Check for dead body
        if (artifact.type == CArtifactType.deadBody && !artifact.takeableStatus)
        {
            cnsl(artifact.name.capitalize() + " is best if left alone.");

            return false;
        }

        // Check for bound monster
        if (artifact.type == CArtifactType.boundMonster)
        {
            cnsl.writeLine(artifact.name.capitalize() + " must be freed.");

            return false;
        }

        if (artifact.weight > 900)
        {
            cnsl.console.writeLine("Don't be absurd.");

            return false;
        }

        if (artifact.weight == -999)
        {
            cnsl.console.writeLine("You can't get that.");

            return false;
        }

        // Check weight exceeds character's max carry limit
        var weightLimit = self.context.gameStateData.runChar.hardiness * 10;

        if (artifact.weight > weightLimit)
        {
            cnsl.writeLine(artifact.name.capitalize() + " is too heavy.");
            return false;
        }

        // Check total encumberance
        if (artifact.weight + self.calculateEncumberance() > weightLimit)
        {
            cnsl.writeLine("You are carrying too much weight. You'll need to drop something else first.");
            return false;
        }

        return true;
    };

    self.printMonsterPersonalityResponse = function(m)
    {
        var response = self.getMonsterName(m);
        var at = " at";

        switch (m.personality)
        {
            case CPersonalityType.friend:
                response += " smile";
                break;

            case CPersonalityType.enemy:
                response += " growl";
                break;

            case CPersonalityType.neutral:
                response += " ignore";
                at = "";
                break;
        }

        if (m.currentGroupSize == 1)
        {
            response += "s";
        }

        response += at;
        response += " you.";

        self.context.console.writeLine(response);
    };

    self.printIntro = function()
    {
        //TODO: Add macro expansion for male/female
        // @XX expands to effect to print for male char
        // $xx expands to effect to print for female char
        self.printDescription(self.context.gameStateData.gameModule.intro);

        self.context.console.writeBlankLine();
    };

    self.registerCommandHandlers = function()
    {
        self.commandHandlers["attack"] = new XCmdHandlerAttack();
        self.commandHandlers["close"] = new XCmdHandlerClose();
        self.commandHandlers["drink"] = new XCmdHandlerDrinkEat(CArtifactType.drinkable);
        self.commandHandlers["drop"] = new XCmdHandlerDrop();
        self.commandHandlers["eat"] = new XCmdHandlerDrinkEat(CArtifactType.edible);
        self.commandHandlers["examine"] = new XCmdHandlerExamine();
        self.commandHandlers["extinguish"] = new XCmdHandlerExtinguish();
        self.commandHandlers["flee"] = new XCmdHandlerFlee();
        self.commandHandlers["free"] = new XCmdHandlerFree();
        self.commandHandlers["get"] = new XCmdHandlerGet();
        self.commandHandlers["take"] = new XCmdHandlerGet();
        self.commandHandlers["give"] = new XCmdHandlerGive();
        self.commandHandlers["hints"] = new XCmdHandlerHints();
        self.commandHandlers["inventory"] = new XCmdHandlerInventory();
        self.commandHandlers["light"] = new XCmdHandlerLight();
        self.commandHandlers["look"] = new XCmdHandlerLook();
        self.commandHandlers["open"] = new XCmdHandlerOpen();
        self.commandHandlers["put"] = new XCmdHandlerPut();
        self.commandHandlers["quit"] = new XCmdHandlerQuit();
        self.commandHandlers["read"] = new XCmdHandlerRead();
        self.commandHandlers["ready"] = new XCmdHandlerReady();
        self.commandHandlers["remove"] = new XCmdHandlerRemove();
        self.commandHandlers["request"] = new XCmdHandlerRequest();
        self.commandHandlers["restore"] = new XCmdHandlerRestore();
        self.commandHandlers["say"] = new XCmdHandlerSay();
        self.commandHandlers["save"] = new XCmdHandlerSave();
        self.commandHandlers["smile"] = new XCmdHandlerSmile();
        self.commandHandlers["status"] = new XCmdHandlerStatus();
        self.commandHandlers["use"] = new XCmdHandlerUse();
        self.commandHandlers["wear"] = new XCmdHandlerWear();

        // Spells
        self.commandHandlers["blast"] = new XCmdHandlerBlast();
        self.commandHandlers["heal"] = new XCmdHandlerHeal();
        self.commandHandlers["power"] = new XCmdHandlerPower();
        self.commandHandlers["speed"] = new XCmdHandlerSpeed();

        // Directional Commands
        var ch = new XCmdHandlerGo(0);
        self.commandHandlers["north"] = ch;
        self.commandHandlers["n"] = ch;
        
        ch = new XCmdHandlerGo(1);
        self.commandHandlers["south"] = ch;
        self.commandHandlers["s"] = ch;

        ch = new XCmdHandlerGo(2);
        self.commandHandlers["east"] = ch;
        self.commandHandlers["e"] = ch;
        
        ch = new XCmdHandlerGo(3);
        self.commandHandlers["west"] = ch;
        self.commandHandlers["w"] = ch;
        
        ch = new XCmdHandlerGo(4);
        self.commandHandlers["up"] = ch;
        self.commandHandlers["u"] = ch;
        
        ch = new XCmdHandlerGo(5);
        self.commandHandlers["down"] = ch;
        self.commandHandlers["d"] = ch;

        // For 6-direction modules
        self.commandHandlers["ne"] = new XCmdHandlerGo(6);
        self.commandHandlers["nw"] = new XCmdHandlerGo(7);
        self.commandHandlers["se"] = new XCmdHandlerGo(8);
        self.commandHandlers["sw"] = new XCmdHandlerGo(9);

        // Debugging commands
        self.commandHandlers["jump"] = new XCmdHandlerJump();
        self.commandHandlers["kill"] = new XCmdHandlerKill();
        self.commandHandlers["acquire"] = new XCmdHandlerAcquire();
    };

    // Convert the raw eamon data file arrays into more readable jscript objects...
    self.initGameModule = function()
    {
        //TODO: Only do this if legacy module, ugly static method call...
        CLegacyConverter.convert(self.context.gameStateData.gameModule);
        
        // Load character's inventory into game module
        self.context.gameStateData.runChar.weaponNumber = 0;

        _.each(self.context.gameStateData.initChar.inventory, function(a)
        {
            var number = _.max(self.context.gameStateData.gameModule.artifacts, function (x) { return x.number; }).number + 1;

            a.number = number;

            self.context.gameStateData.gameModule.artifacts.push(a);

            if ((a.type == CArtifactType.weapon || a.type == CArtifactType.magicWeapon) && self.context.gameStateData.runChar.weaponNumber <= 0)
            {
                self.context.gameStateData.runChar.weaponNumber = number;
            }
        });
    };
};